import pandas as pd
import torch


def prediction(model, images, targets=None):
    """
    Makes and stores predictions on one batch of images.
    :param model: The model used for prediction.
    :param images: The batch of images on which to predict bounding boxes.
    :param targets: Optional ground truth labels and bounding boxes.
    :return: A Dataframe consisting of predicted and (if supplied) ground truth labels and bounding boxes.
    """
    predictions = {"pred_boxes": [],
                   "pred_scores": [],
                   "pred_labels": []}
    if targets:
        predictions["gt_boxes"] = [target["bboxes"].detach().cpu().numpy() for target in targets]
        predictions["gt_labels"] = [target["labels"].detach().cpu().numpy() for target in targets]
    with torch.no_grad():
        det = model(images)
        for i in range(images.shape[0]):
            pred_det = det[i]
            predictions["pred_boxes"].append(pred_det[:, :4].detach().cpu().numpy())
            predictions["pred_scores"].append(pred_det[:, 4].detach().cpu().numpy())
            predictions["pred_labels"].append(pred_det[:, 5].detach().cpu().numpy().astype(int))
    return predictions


def prediction_df(model, test_loader,
                  device=torch.device("cuda" if torch.cuda.is_available() else "cpu"),
                  verbose=0):
    """
    Creates a Dataframe of predictions from unlabeled test data.
    :param model: The model used for predictions.
    :param test_loader: Dataloader for unlabeled test data.
    :param device: Device to load the model on. CPU or CUDA. Defaults using CUDA if available, otherwise CPU.
    :param verbose: If positive prints output every "verbose" steps.
    :return: Dataframe of bounding box predictions.
    """
    assert verbose >= 0
    predictions = []
    n = 0
    for images in test_loader:
        n += 1
        images = torch.stack(images).to(device).float()
        predictions.append(pd.DataFrame(prediction(model, images)))
        if verbose and n % verbose == 0:
            print(f"Processed batch {n}.")

    return pd.concat(predictions).reset_index(drop=True)


def val_prediction_df(model, test_loader,
                      device=torch.device("cuda" if torch.cuda.is_available() else "cpu"),
                      verbose=0):
    """
    Creates a Dataframe of predictions from labeled data.
    :param model: The model used for predictions.
    :param test_loader: Dataloader for unlabeled test data.
    :param device: Device to load the model on. CPU or CUDA. Defaults using CUDA if available, otherwise CPU.
    :param verbose: If positive prints output every "verbose" steps.
    :return: Dataframe of predicted and ground truth bounding box labels.
    """
    assert verbose >= 0
    predictions = []
    n = 0
    for images, labels in test_loader:
        n += 1
        images = torch.stack(images).to(device).float()
        predictions.append(pd.DataFrame(prediction(model, images, labels)))
        if verbose and n % verbose == 0:
            print(f"Processed batch {n}.")

    return pd.concat(predictions).reset_index(drop=True)



import torch


def load_scheduler(scheduler_type,
                   optimizer,
                   verbose=True,
                   **kwargs):
    """
    Loads the selected learning rate scheduler for training.
    :param scheduler_type: String for the scheduler type
    :param optimizer: Optimizer for model training.
    :param verbose: Print message to stdout for each update. Defaults to True.
    :param kwargs: Any scheduler specific variables required.
    :return: The learning rate scheduler for training.
    """
    if scheduler_type.lower() == "cosine":
        return torch.optim.lr_scheduler.CosineAnnealingLR(optimizer=optimizer,
                                                          T_max=kwargs.get("T_max", 100),
                                                          eta_min=kwargs.get("eta_min", 0),
                                                          verbose=verbose)
    elif scheduler_type.lower() == "cosinewr":
        return torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer=optimizer,
                                                                    T_0=kwargs["T_0"],
                                                                    T_mult=kwargs.get("T_mult", 1),
                                                                    eta_min=kwargs.get("eta_min", 0),
                                                                    verbose=verbose)
    elif scheduler_type.lower() == "exponential":
        return torch.optim.lr_scheduler.ExponentialLR(optimizer=optimizer,
                                                      gamma=kwargs.get("gamma", 1.0),
                                                      verbose=verbose)
    elif scheduler_type.lower() == "mult":
        return torch.optim.lr_scheduler.MultiplicativeLR(optimizer=optimizer,
                                                         lr_lambda=kwargs.get("lr_lambda"),
                                                         verbose=verbose)
    elif scheduler_type.lower() == "reduce":
        return torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer=optimizer,
                                                          mode=kwargs["mode"],
                                                          factor=kwargs.get("factor", 0.1),
                                                          patience=kwargs.get("patience", 10),
                                                          threshold=kwargs.get("threshold", 1e-4),
                                                          threshold_mode=kwargs.get("threshold_mode", "rel"),
                                                          cooldown=kwargs.get("cooldown", 0),
                                                          min_lr=kwargs.get("min_lr", 0),
                                                          verbose=verbose)



import torch
from .petrel import PetrelDataset


class ValDataset(PetrelDataset):
    """
    Dataset class for validation data. Performs one preprocessing transformation on the raw data.
    """

    def __init__(self, meta_data,
                 boxes,
                 image_root,
                 transform=None,
                 train_pipe=False):
        """
        :param meta_data: DataFrame consisting of metadata for dataset. Must be one row per datapoint.
        :param boxes: DataFrame consisting of all bounding box data for dataset. Must be in PASCAL VOC format, with
        xmin, ymin, xmax, ymax in separate columns. Must also include numeric label column.
        :param image_root: Root directory for imge files.
        :param transform: Any preprocessing transformations used in preparing the data. Usually Albumentations.
        :param train_pipe: Is the dataset being used in a training pipeline. Needed to determine
        """
        super(ValDataset, self).__init__(meta_data, boxes, image_root, transform, train_pipe)

    def __getitem__(self, index):
        """
        Retrieves the image and boxes with the specified index.
        :param index: Index of the meta_data row for the item to be loaded.
        :return: Image and target dictionary consisting of bounding boxes and labels.
        """

        image_meta = self.meta_data.loc[index]
        image_boxes = self.boxes[self.boxes["file"] == image_meta["file"]]
        image, bboxes, labels = self.load_image_and_boxes(image_meta, image_boxes)
        target = {"bboxes": bboxes,
                  "labels": labels}

        if self.transform:
            sample = self.transform(image=image,
                                    bboxes=target["bboxes"],
                                    labels=target["labels"])
            image, target["bboxes"] = sample['image'], self._box_to_tensor(sample["bboxes"])
            target["labels"] = torch.tensor(sample["labels"])
        return image, target



from . import AvgCounter


class LossCounter:
    """
    Stores counters for total loss, class loss and bounding box loss.
    """

    def __init__(self):
        self._loss = AvgCounter()
        self._class_loss = AvgCounter()
        self._box_loss = AvgCounter()

    @property
    def avg(self):
        """
        Returns the average value of the main loss counter.
        :return: The average value of the main loss counter.
        """
        return self._loss.avg

    @property
    def box_avg(self):
        """
        Return the average value of the bounding box loss counter.
        :return: The average value of the bounding box loss counter.
        """
        return self._box_loss.avg

    @property
    def box_loss(self):
        """
        Returns the bounding box loss counter.
        :return: The bounding box loss counter.
        """
        return self._box_loss

    @property
    def class_avg(self):
        """
        Returns the average value of the class loss counter.
        :return: The average value of the class loss counter.
        """
        return self._class_loss.avg

    @property
    def class_loss(self):
        """
        Returns the class loss counter.
        :return: The class loss counter
        """
        return self._class_loss

    @property
    def loss(self):
        """
        Returns the total loss counter.
        :return: The total loss counter.
        """
        return self._loss

    def update(self, output, batch_size=1):
        """
        Updates the loss counters after a forward pass of a batch of data. Assumes the update values
        are the sums of the update values for the individual batches.
        :param output: The output of the forwrd pass.
        :param batch_size: The batch size of the forward pass.
        """
        self._loss.update(output["loss"], batch_size)
        self._class_loss.update(output["class_loss"], batch_size)
        self._box_loss.update(output["box_loss"], batch_size)

    def update_avg(self, output, batch_size=1):
        """
        Updates the loss counters after a forward pass of a batch of data. Assumes the update values
        are the averages of the update values for the individual batches.
        :param output: The output of the forwrd pass.
        :param batch_size: The batch size of the forward pass.
        """
        self._loss.update_avg(output["loss"], batch_size)
        self._class_loss.update_avg(output["class_loss"], batch_size)
        self._box_loss.update_avg(output["box_loss"], batch_size)


from unittest import TestCase
from petrel.model import LossCounter


class TestLoss(TestCase):
    def test_counter(self):
        loss = LossCounter()
        loss.update({"loss": 5, "class_loss": 2, "box_loss": 1}, batch_size=1)
        loss.update({"loss": 10, "class_loss": 4, "box_loss": 2}, batch_size=2)
        assert loss.avg == 5.0
        assert loss.class_avg == 2.0
        assert loss.box_avg == 1.0
        loss.update({"loss": 19, "class_loss": 34, "box_loss": 16}, batch_size=2)
        assert loss.avg == 6.8
        assert loss.class_avg == 8.0
        assert loss.box_avg == 3.8

    def test_avg_counter(self):
        loss = LossCounter()
        loss.update_avg({"loss": 5, "class_loss": 2, "box_loss": 1}, batch_size=1)
        loss.update_avg({"loss": 5, "class_loss": 2, "box_loss": 1}, batch_size=2)
        assert loss.avg == 5.0
        assert loss.class_avg == 2.0
        assert loss.box_avg == 1.0
        loss.update_avg({"loss": 19, "class_loss": 34, "box_loss": 16}, batch_size=2)
        assert loss.avg == 10.6
        assert loss.class_avg == 14.8
        assert loss.box_avg == 7.0



import numpy as np
import pandas as pd
from effdet.evaluation.detection_evaluator import PascalDetectionEvaluator


def model_eval(pred, categories, ious=np.round(np.arange(0.5, 1.0, 0.05), 2)):
    """
    Evaluates mAP scores at various IOU thresholds. Returns DataFrame of mAP scores for each threshold,
    label and the unweighted average of label mAPs.
    :param pred: Dataframe of predicted and ground truth labels and bounding boxes.
    :param categories: List of category names and labels. List elements must take the form {"id": id, "name": name}
    :param ious: IOU thresholds for computing mAPs. Defaults to .50:.05:.95.
    :return: Dataframe of mAP values.
    """
    evaluators = {iou: PascalDetectionEvaluator(categories=categories,
                                                matching_iou_threshold=iou) for iou in ious}
    for n, row in pred.iterrows():
        gt = {"cls": row["gt_labels"],
              "bbox": row["gt_boxes"]}
        det = {"cls": row["pred_labels"],
               "bbox": row["pred_boxes"],
               "score": row["pred_scores"]}
        for iou in evaluators:
            evaluators[iou].add_single_ground_truth_image_info(n, gt)
            evaluators[iou].add_single_detected_image_info(n, det)

    evaluators = {iou: evaluators[iou].evaluate() for iou in evaluators}
    for iou in evaluators:
        new_map = {}
        for k in evaluators[iou]:
            if "Precision" in k:
                new_map["Precision"] = evaluators[iou][k]
            else:
                new_k = k.split("/")[-1]
                new_map[new_k] = evaluators[iou][k]
        evaluators[iou] = new_map
    mAP = pd.DataFrame(evaluators).T
    return mAP



from petrel.model import load_edet, load_optimizer
from torch.optim import Adadelta, Adagrad, Adam, AdamW, Adamax, RMSprop, Rprop, SGD

from unittest import TestCase


class TestOptimizer(TestCase):
    def test_adadelta(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("adadelta", model, learning_rate=1e-2)
        assert optimizer.__class__ == Adadelta
        assert optimizer.param_groups[0]["lr"] == 1e-2
        assert optimizer.param_groups[0]["weight_decay"] == 0

    def test_adagrad(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("adagrad", model, learning_rate=1e-2)
        assert optimizer.__class__ == Adagrad
        assert optimizer.param_groups[0]["lr"] == 1e-2
        assert optimizer.param_groups[0]["weight_decay"] == 0

    def test_adam(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("adam", model, learning_rate=1e-2)
        assert optimizer.__class__ == Adam
        assert optimizer.param_groups[0]["lr"] == 1e-2
        assert optimizer.param_groups[0]["weight_decay"] == 0
        assert not optimizer.param_groups[0]["amsgrad"]

    def test_adamw(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("adamw", model, learning_rate=1e-2)
        assert optimizer.__class__ == AdamW
        assert optimizer.param_groups[0]["lr"] == 1e-2
        assert optimizer.param_groups[0]["weight_decay"] == 4e-5
        assert not optimizer.param_groups[0]["amsgrad"]

    def test_adammax(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("adamax", model, learning_rate=1e-2)
        assert optimizer.__class__ == Adamax
        assert optimizer.param_groups[0]["lr"] == 1e-2
        assert optimizer.param_groups[0]["weight_decay"] == 0

    def test_rmsprop(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("rmsprop", model, learning_rate=1e-2)
        assert optimizer.__class__ == RMSprop
        assert optimizer.param_groups[0]["lr"] == 1e-2
        assert optimizer.param_groups[0]['momentum'] == 0.9
        assert optimizer.param_groups[0]["weight_decay"] == 0

    def test_rprop(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("rprop", model, learning_rate=1e-2)
        assert optimizer.__class__ == Rprop
        assert optimizer.param_groups[0]["lr"] == 1e-2

    def test_sgd(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("sgd", model, learning_rate=1e-2)
        assert optimizer.__class__ == SGD
        assert optimizer.param_groups[0]["lr"] == 1e-2
        assert optimizer.param_groups[0]['momentum'] == 0.9
        assert optimizer.param_groups[0]["weight_decay"] == 0



class AvgCounter:
    """
    Stores totals and counts.
    """

    def __init__(self):
        self._sum = 0
        self._count = 0

    def reset(self):
        self._sum = 0
        self._count = 0

    @property
    def avg(self):
        """
        Returns the current average.
        :return: The current average.
        """
        return self._sum/self._count

    def concat(self, other_counter):
        """
        Add values from anther counter.
        :param other_counter: Second counter whose totals are added to the current counter.
        """
        self._sum += other_counter.sum()
        self._count += other_counter.count()

    @property
    def count(self):
        """
        Returns the current count
        :return: The current count.
        """
        return self._count

    @property
    def sum(self):
        """
        Returns the current total.
        :return: The current total.
        """
        return self._sum

    def update(self, val, count):
        """
        Updates totals and counts. Assumes val is the sum of count elements
        :param val: New value to add to the total
        :param count: Count of new elements being added.
        """
        self._sum += val
        self._count += count

    def update_avg(self, avg, count):
        """
        Updates totals and counts. Assumes avg is the average of the values of count elements
        :param avg: Average of new values being added.
        :param count: Count of new elements being added.
        """
        self._sum += avg * count
        self._count += count





from unittest import TestCase
from petrel.model import AvgCounter


class TestCounter(TestCase):
    def test_counter(self):
        avg_counter = AvgCounter()
        avg_counter.update(2, 1)
        avg_counter.update(3, 1)
        avg_counter.update(1, 1)
        assert avg_counter.avg == 2.0
        assert avg_counter.sum == 6.0
        assert avg_counter.count == 3
        avg_counter.update(2, 1)
        assert avg_counter.avg == 2.0
        assert avg_counter.sum == 8.0
        assert avg_counter.count == 4
        avg_counter.update(7, 1)
        assert avg_counter.avg == 3.0
        assert avg_counter.sum == 15.0
        assert avg_counter.count == 5
        avg_counter.reset()
        assert avg_counter.sum == 0.0
        assert avg_counter.count == 0

    def test_avg_counter(self):
        avg_counter = AvgCounter()
        avg_counter.update_avg(2, 2)
        assert avg_counter.avg == 2.0
        assert avg_counter.sum == 4.0
        assert avg_counter.count == 2
        avg_counter.update_avg(3, 3)
        assert avg_counter.avg == 2.6
        assert avg_counter.sum == 13.0
        assert avg_counter.count == 5
        avg_counter.update_avg(8, 1)
        assert avg_counter.avg == 3.5
        assert avg_counter.sum == 21.0
        assert avg_counter.count == 6

    def test_variable_count(self):
        avg_counter = AvgCounter()
        avg_counter.update(2, 2)
        assert avg_counter.avg == 1.0
        assert avg_counter.sum == 2.0
        assert avg_counter.count == 2
        avg_counter.update(3, 3)
        assert avg_counter.avg == 1.0
        assert avg_counter.sum == 5.0
        assert avg_counter.count == 5
        avg_counter.update(7, 1)
        assert avg_counter.avg == 2.0
        assert avg_counter.sum == 12.0
        assert avg_counter.count == 6



from effdet import DetBenchTrain


class DetBenchTrainVal(DetBenchTrain):
    """
    Wrapper class around DetBenchTrain to override the forward method. DetBenchTrain computes detection on
    validation, which slows down the validation portion of the a training pipeline and is not always desirable.
    This module allows validation to be evaluated purely by the loss functions.
    """
    def __init__(self, model, config):
        """
        :param model: EfficientDet model to train.
        :param config: Model configuration dictionary.
        """
        super(DetBenchTrainVal, self).__init__(model, config)

    def forward(self, x, target):
        """
        Performs forward propagation on a batch.
        :param x: Batch of images.
        :param target: Batch of bounding box labels.
        :return: Output dictionary consisting of total loss, class loss, and box loss.
        """
        class_out, box_out = self.model(x)
        if self.anchor_labeler is None:
            # target should contain pre-computed anchor labels if labeler not present in bench
            assert 'label_num_positives' in target
            cls_targets = [target[f'label_cls_{l}'] for l in range(self.num_levels)]
            box_targets = [target[f'label_bbox_{l}'] for l in range(self.num_levels)]
            num_positives = target['label_num_positives']
        else:
            cls_targets, box_targets, num_positives = self.anchor_labeler.batch_label_anchors(
                target['bbox'], target['cls'])

        loss, class_loss, box_loss = self.loss_fn(class_out, box_out, cls_targets, box_targets, num_positives)
        output = {'loss': loss, 'class_loss': class_loss, 'box_loss': box_loss}
        return output



from abc import ABC

import numpy as np
import cv2
import torch
from torch.utils.data import Dataset


class PetrelDataset(Dataset, ABC):
    """Dataset class for data used in Petrel models."""
    def __init__(self, meta_data,
                 boxes,
                 image_root,
                 transform=None,
                 train_pipe=False):
        """
        :param meta_data: DataFrame consisting of metadata for dataset. Must be one row per datapoint.
        :param boxes: DataFrame consisting of all bounding box data for dataset. Must be in PASCAL VOC format, with
        xmin, ymin, xmax, ymax in separate columns. Must also include numeric label column.
        :param image_root: Root directory for imge files.
        :param transform: Any preprocessing transformations used in preparing the data. Usually Albumentations.
        :param train_pipe: Is ths dataset used for training prediction. Training requires yxyx while prediction
        requires xyxy. Defaults to False.
        """
        super(PetrelDataset).__init__()

        self.meta_data = meta_data
        self.boxes = boxes
        self.image_root = image_root
        self.transform = transform
        self.train_pipe = train_pipe

    def _box_to_tensor(self, boxes):
        """
        Converts bounding box data to PyTorch tensor.
        :param boxes: List of bounding boxes.
        :return: Bounding box tensor.
        """
        bboxes = torch.tensor(boxes) if len(boxes) > 0 else torch.zeros((0, 4))
        # Convert bounded box to yxyx format if in training pipline
        if self.train_pipe:
            bboxes[:, [0, 1, 2, 3]] = bboxes[:, [1, 0, 3, 2]]
        return bboxes

    def __len__(self) :
        """
        Returns the size of the dataset.
        :return: Size of the dataset.
        """
        return self.meta_data.shape[0]

    def load_image_and_boxes(self, image_meta, image_boxes):
        """
        Loads single image with all bounding boxes and associated labels
        :param image_meta: DataFrame row consisting of data for the single image.
        :param image_boxes: DataFrame of bounding boxes and labels for the image.
        :return: Image, bounding boxes, and labels.
        """
        image = cv2.cvtColor(
            cv2.imread(f"{self.image_root}/{image_meta['file']}",
                       cv2.IMREAD_COLOR),
            cv2.COLOR_BGR2RGB).astype(np.float32) / 255.0
        bboxes = image_boxes[["xmin", "ymin", "xmax", "ymax"]].values
        bboxes = torch.tensor(bboxes)
        if "labels" in image_boxes.columns:
            labels = torch.tensor(image_boxes["labels"].values,
                                  dtype=torch.int64)
        else:
            labels = torch.tensor(np.ones(bboxes.shape[0]),
                                  dtype=torch.int64)
        return image, bboxes, labels



import os

import numpy as np
import torch
import time
from .loss import LossCounter


class ModelTrainer:
    """
    Class to train object detection models
    """

    def __init__(self, model,
                 optimizer,
                 scheduler,
                 base_dir,
                 device=torch.device("cuda" if torch.cuda.is_available() else "cpu"),
                 num_epochs=100,
                 model_file=None,
                 log_path=None,
                 verbose=True,
                 verbose_step=100,
                 keep_models=3):
        """

        :param model: Model to train.
        :param device: Device to train the model on. CPU or CUDA.
        :param optimizer: Model optimizer.
        :param scheduler: Learning rate scheduler for training.
        :param base_dir: Default directory for saving output.
        :param num_epochs: Number of epochs to train. Defaults to 100
        :param model_file: File for loading partially trained models.
        :param log_path: Optional value for specifying log file location
        :param verbose: Boolean for printing output. Defaults to True
        :param verbose_step: Interval on which to print status. Defaults to 100.
        :param keep_models: Maximum number of models to save. Defaults to 3
        """
        self.model = model
        self.device = device
        self.model.to(self.device)
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.num_epochs = num_epochs
        self.base_dir = base_dir
        self.verbose = verbose
        self.verbose_step = verbose_step
        self.keep_models = keep_models
        self.zfill = np.ceil(np.log10(self.num_epochs)).astype(int)
        if model_file:
            self.load(model_file)
        else:
            self.start_epoch = 0
            self.best_summary_loss = []
        if not os.path.exists(self.base_dir):
            os.makedirs(self.base_dir)

        self.log_path = log_path if log_path else f"{self.base_dir}/log.csv"
        print(f'Fitter prepared. Device is {self.device}')

    def _csv_header(self):
        """
        Writes log file header.
        """
        self.log("Epoch, Stage, Summary Loss, Class Loss, Box Loss", print_line=False)

    @staticmethod
    def _csv_line(summary_loss, epoch, stage):
        """
        Builds the string for one line of logging to csv file.

        :param summary_loss: Total current loss
        :param epoch: Current epoch
        :param stage: Training or Validation
        :return: String to write to log file.
        """
        return f"{epoch}, {stage}, {summary_loss.avg:.5f}," + \
               f"{summary_loss.class_avg:.5f}," + \
               f"{summary_loss.box_avg:.5f}"

    @staticmethod
    def _output_dict(output):
        return {"loss": output["loss"].detach().item(),
                "class_loss": output["class_loss"].detach().item(),
                "box_loss": output["box_loss"].detach().item()}

    @staticmethod
    def _print_line(summary_loss, step, total_steps, stage, t):
        """
        Prints a line of output showing the current status of training.

        :param summary_loss: Total current loss.
        :param step: Current step in epoch.
        :param total_steps: Total number of steps in epoch.
        :param stage: Training or Validation.
        :param t: Start time of training or validation epoch.
        """
        print(
            f"{stage} Step {step}/{total_steps}, " +
            f"summary_loss: {summary_loss.avg:.5f}, " +
            f"class_loss: {summary_loss.class_avg:.5f}, " +
            f"box_loss: {summary_loss.box_avg:.5f}, " +
            f"time: {(time.time() - t):.5f}\n")

    def fit(self, train_loader, validation_loader):
        """
        Trains the model. Saves best models after each epoch and updates learning rate.

        :param train_loader: Data Loader for training data.
        :param validation_loader: Data Loader for validation data.
        """
        if self.start_epoch == 0:
            self._csv_header()
        for epoch in range(self.start_epoch, self.num_epochs):
            # Train one epoch
            summary_loss = self.train(train_loader)

            self.log(self._csv_line(summary_loss, epoch, "Train"))
            self.save(f'{self.base_dir}/last-checkpoint.bin', epoch)

            # Evaluate on validation data.
            summary_loss = self.validate(validation_loader)

            self.log(self._csv_line(summary_loss, epoch, "Val"))

            # Update list of saved models
            if len(self.best_summary_loss) < self.keep_models:
                self.best_summary_loss.append((summary_loss.avg, epoch))
                self.save(f'{self.base_dir}/best-checkpoint-{str(epoch).zfill(self.zfill)}epoch.bin', epoch)
                self.best_summary_loss.sort()
            elif summary_loss.avg < self.best_summary_loss[-1][0]:
                _, old_epoch = self.best_summary_loss.pop()
                self.best_summary_loss.append((summary_loss.avg, epoch))
                self.save(f'{self.base_dir}/best-checkpoint-{str(epoch).zfill(self.zfill)}epoch.bin', epoch)
                os.remove(f'{self.base_dir}/best-checkpoint-{str(old_epoch).zfill(self.zfill)}epoch.bin')
                self.best_summary_loss.sort()

            # Update learning rate.
            if self.scheduler:
                self.scheduler.step()

    def load(self, path):
        """
        Loads saved model, including optimizer, scheduler history, and previous best models.

        :param path: Path of model to load.
        """
        checkpoint = torch.load(path)
        self.model.model.load_state_dict(checkpoint["state_dict"])
        self.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
        self.scheduler.load_state_dict(checkpoint["scheduler_state_dict"])
        self.best_summary_loss = checkpoint["best_summary_loss"]
        self.start_epoch = checkpoint["epoch"] + 1

    def log(self, message, print_line=True):
        """
        Write data to log.
        :param message: Message to write.
        :param print_line: Boolean, print message to stdout. Default True.
        """
        if self.verbose and print_line:
            print(message)
        with open(self.log_path, 'a+') as logger:
            logger.write(f'{message}\n')

    def save(self, path, epoch):
        """
        Saves model after training epoch.

        :param path: Path of file for saved model
        :param epoch: Epoch number.
        """
        self.model.eval()
        torch.save({
            "state_dict": self.model.model.state_dict(),
            "optimizer_state_dict": self.optimizer.state_dict(),
            "scheduler_state_dict": self.scheduler.state_dict(),
            "best_summary_loss": self.best_summary_loss,
            "epoch": epoch,
        }, path)

    def train(self, train_loader):
        """
        Trains model for one epoch.
        :param train_loader: Data Loader for training data
        :type train_loader: torch.utils.data.DataLoader
        :return: Total Loss on training data for epoch.
        """
        self.model.train()
        summary_loss = LossCounter()
        t = time.time()
        for step, (images, targets) in enumerate(train_loader):
            if self.verbose and step and step % self.verbose_step == 0:
                self._print_line(summary_loss, step, len(train_loader), "Train", t)

            images = torch.stack(images)
            images = images.to(self.device).float()
            batch_size = images.shape[0]
            boxes = [target['bboxes'].to(self.device).float() for target in targets]
            labels = [target['labels'].to(self.device).float() for target in targets]

            self.optimizer.zero_grad()

            output = self.model(images, {'bbox': boxes, 'cls': labels})

            output['loss'].backward()

            summary_loss.update_avg(self._output_dict(output), batch_size)

            self.optimizer.step()

        return summary_loss

    def validate(self, val_loader):
        """
        Evaluates model on validation data after one epoch of training.

        :param val_loader: Data Loader for validation data.
        :type val_loader: torch.utils.data.DataLoader
        :return Total Loss on validation data.
        """
        self.model.eval()
        summary_loss = LossCounter()
        t = time.time()
        for step, (images, targets) in enumerate(val_loader):
            if self.verbose and step and step % self.verbose_step == 0:
                self._print_line(summary_loss, step, len(val_loader), "Val", t)
            with torch.no_grad():
                images = torch.stack(images)
                batch_size = images.shape[0]
                images = images.to(self.device).float()
                boxes = [target['bboxes'].to(self.device).float() for target in targets]
                labels = [target['labels'].to(self.device).float() for target in targets]

                output = self.model(images, {'bbox': boxes, 'cls': labels,
                                             "img_scale": None,
                                             "img_size": None})
                summary_loss.update_avg(self._output_dict(output), batch_size)

        return summary_loss



import torch
from .petrel import PetrelDataset


class TrainDataset(PetrelDataset):
    """
    Dataset class for training data. Performs the preprocessing transformation up to max_iter times, stopping if
    there is at least one bounding box in the transformed image. Performs another preprocessing transformation
    if the first cannot produce bounding boxes.
    """

    def __init__(self, meta_data,
                 boxes,
                 image_root,
                 transform,
                 backup_transform=None,
                 max_iter=100):
        """
        :param meta_data: DataFrame consisting of metadata for dataset. Must be one row per datapoint.
        :param boxes: DataFrame consisting of all bounding box data for dataset. Must be in PASCAL VOC format, with
        xmin, ymin, xmax, ymax in separate columns. Must also include numeric label column.
        :param image_root: Root directory for imge files.
        :param transform: Any preprocessing transformations used in preparing the data. Usually Albumentations.
        :param backup_transform: Optional second transform to use if the first does not produce output with bounding boxes.
        :param max_iter: Maximum number of times to try transform.
        """
        super(TrainDataset, self).__init__(meta_data, boxes, image_root, transform, train_pipe=True)
        self.backup_transform = backup_transform if backup_transform else transform
        self.max_iter = max_iter

    def __getitem__(self, index):
        """
        Retrieves the image and boxes with the specified index.
        :param index: Index of the meta_data row for the item to be loaded.
        :return: Image and target dictionary consisting of bounding boxes and labels.
        """

        image_meta = self.meta_data.loc[index]
        image_boxes = self.boxes[self.boxes["file"] == image_meta["file"]]
        image, bboxes, labels = self.load_image_and_boxes(image_meta, image_boxes)
        target = {"bboxes": bboxes,
                  "labels": labels}
        # Try standard transform in bounding boxes present in raw image.
        if self.transform and bboxes.shape[0] > 0:
            # Try the standard transformation up to max_iter times.
            for _ in range(self.max_iter):
                sample = self.transform(image=image,
                                        bboxes=target["bboxes"],
                                        labels=target["labels"])
                ## Return preprocessed image if bounding boxes present.
                if len(sample["bboxes"]) > 0:
                    image, target["bboxes"] = sample["image"], self._box_to_tensor(sample["bboxes"])
                    target["labels"] = torch.tensor(sample["labels"])
                    return image, target
        # Apply the backup transform if the standard transform failed.
        sample = self.backup_transform(image=image,
                                       bboxes=target["bboxes"],
                                       labels=target["labels"])
        image, target["bboxes"] = sample["image"], self._box_to_tensor(sample["bboxes"])
        target["labels"] = torch.tensor(sample["labels"])
        return image, target



import albumentations as A
from albumentations.pytorch.transforms import ToTensorV2

from .petrel import PetrelDataset
from .train_data import TrainDataset
from .val_data import ValDataset

BBOX = A.BboxParams(
    format='pascal_voc',
    min_area=0,
    min_visibility=0,
    label_fields=['labels'])


def get_default_transform(img_size):
    """Returns a function to perform the default transform if the training
       transform fails.
    """
    return A.Compose([A.Resize(height=img_size[0],
                               width=img_size[1], p=1.0),
                      ToTensorV2(p=1.0)],
                     bbox_params=BBOX,
                     p=1.0)



from effdet.bench import DetBenchPredict, DetBenchTrain
from unittest import TestCase

from petrel.model import load_edet, DetBenchTrainVal


class TestEnet(TestCase):
    def test_classes(self):
        model = load_edet(config_name="tf_efficientdet_d0",
                          image_size=(512, 512),
                          num_classes=100)
        assert model.num_classes == 100
        assert model.num_levels == 5
        assert model.max_det_per_image == 1000
        assert model.max_detection_points == 5000
        assert not model.soft_nms
        assert model.__class__ == DetBenchTrainVal

    def test_det_per_image(self):
        model = load_edet(config_name="tf_efficientdet_d0",
                          image_size=(512, 512),
                          max_det_per_image=100)
        assert model.num_classes == 1
        assert model.num_levels == 5
        assert model.max_det_per_image == 100
        assert model.max_detection_points == 5000
        assert not model.soft_nms
        assert model.__class__ == DetBenchTrainVal

    def test_enet(self):
        model = load_edet(config_name="tf_efficientdet_d0",
                          image_size=(512, 512))
        assert model.num_classes == 1
        assert model.num_levels == 5
        assert model.max_det_per_image == 1000
        assert model.max_detection_points == 5000
        assert not model.soft_nms
        assert model.__class__ == DetBenchTrainVal

    def test_image_int(self):
        model = load_edet(config_name="tf_efficientdet_d0",
                          image_size=512,
                          num_classes=100)
        assert model.num_classes == 100
        assert model.num_levels == 5
        assert model.max_det_per_image == 1000
        assert model.max_detection_points == 5000
        assert not model.soft_nms
        assert model.__class__ == DetBenchTrainVal
        assert model.config.image_size == [512, 512]

    def test_image_size(self):
        model = load_edet(config_name="tf_efficientdet_d0",
                          image_size=(256, 256),
                          num_classes=100)
        assert model.num_classes == 100
        assert model.num_levels == 5
        assert model.max_det_per_image == 1000
        assert model.max_detection_points == 5000
        assert not model.soft_nms
        assert model.__class__ == DetBenchTrainVal
        assert model.config.image_size == [256, 256]

    def test_soft_nms(self):
        model = load_edet(config_name="tf_efficientdet_d0",
                          image_size=(512, 512),
                          soft_nms=True)
        assert model.num_classes == 1
        assert model.num_levels == 5
        assert model.max_det_per_image == 1000
        assert model.max_detection_points == 5000
        assert model.soft_nms
        assert model.__class__ == DetBenchTrainVal

    def test_train(self):
        model = load_edet(config_name="tf_efficientdet_d0",
                          image_size=(512, 512),
                          train=False)
        assert model.num_classes == 1
        assert model.num_levels == 5
        assert model.max_det_per_image == 1000
        assert model.max_detection_points == 5000
        assert not model.soft_nms
        assert model.__class__ == DetBenchPredict

    def test_checkout(self):
        model = load_edet(config_name="tf_efficientdet_d1",
                          image_size=(640, 640),
                          num_classes=90,
                          checkpoint_path="tests/tf_efficientdet_d1_40-a30f94af.pth",
                          train=True)
        assert model.num_classes == 90
        assert model.num_levels == 5
        assert model.max_det_per_image == 1000
        assert model.max_detection_points == 5000
        assert not model.soft_nms
        assert model.__class__ == DetBenchTrainVal

    def test_detection(self):
        model = load_edet(config_name="tf_efficientdet_d0",
                          image_size=(512, 512),
                          detection=True)
        assert model.num_classes == 1
        assert model.num_levels == 5
        assert model.max_det_per_image == 1000
        assert model.max_detection_points == 5000
        assert not model.soft_nms
        assert model.__class__ == DetBenchTrain



from .bench import DetBenchTrainVal
from .counter import AvgCounter
from .edet import load_edet
from .loss import LossCounter
from .optimizer import load_optimizer
from .scheduler import load_scheduler
from .trainer import ModelTrainer



import torch
from effdet import get_efficientdet_config, create_model_from_config
from .bench import DetBenchTrainVal


def load_edet(config_name,
              image_size,
              checkpoint_path=None,
              num_classes=1,
              max_det_per_image=1000,
              soft_nms=False,
              device=torch.device("cuda" if torch.cuda.is_available() else "cpu"),
              train=True,
              detection=False):
    """
    Loads the EfficientDet model with the given config name, input size, and output classes.
    Can load models with COCO pretrained weights or user-defined model weights.


    :param config_name: Name of the mdoel to load.
    :param image_size: Size the input images.
    :param checkpoint_path: Option path for loading model weights.
    :param num_classes: Number of prediction classes.
    :param max_det_per_image: Maximum number of detection regions to predict.
    :param soft_nms: Use soft non-max suppression. Defaults to False as soft nms is very slow.
    :param device: Device to load the model on. CPU or CUDA. Defaults using CUDA if available, otherwise CPU
    :param train: Is the model being trained. Defaults to True
    :param detection: Return detections during training. This slows down the validation step. Defaults to False.
    :return: EfficientDet model.
    """
    config = get_efficientdet_config(config_name)
    if type(image_size) == int:
        config.image_size = [image_size, image_size]
    else:
        config.image_size = image_size
    config.max_det_per_image = max_det_per_image
    config.soft_nms = soft_nms
    pretrained = False if checkpoint_path else True
    model = create_model_from_config(config, pretrained=pretrained,
                                     bench_task="train" if train else "predict",
                                     max_det_per_image=max_det_per_image,
                                     num_classes=num_classes,
                                     soft_nms=soft_nms,
                                     checkpoint_path=checkpoint_path if checkpoint_path else "",
                                     bench_labeler=True)

    if train:
        if not detection:
            model = DetBenchTrainVal(model.model, model.config)
        model.train()
    else:
        model.eval()
    if device:
        model = model.to(device)

    return model



import torch


def load_optimizer(optimizer_type,
                   model,
                   learning_rate,
                   **kwargs):
    """
    Returns the optimizer for model training.
    :param optimizer_type: Returns the
    :param model: The model to be trained.
    :param learning_rate: Initial learning rate.
    :param kwargs:
    :return: Optimizer for model training.
    """
    if optimizer_type.lower() == "adadelta":
        return torch.optim.Adadelta(model.parameters(),
                                    lr=learning_rate,
                                    rho=kwargs.get("rho", 0.9),
                                    weight_decay=kwargs.get("weight_decay", 0.0))
    elif optimizer_type.lower() == "adagrad":
        return torch.optim.Adagrad(model.parameters(),
                                   lr=learning_rate,
                                   lr_decay=kwargs.get("lr_decay", 0),
                                   weight_decay=kwargs.get("weight_decay", 0))
    elif optimizer_type.lower() == "adam":
        return torch.optim.Adam(model.parameters(),
                                lr=learning_rate)
    elif optimizer_type.lower() == "adamw":
        return torch.optim.AdamW(model.parameters(),
                                 lr=learning_rate,
                                 weight_decay=kwargs.get("weight_decay", 4e-5))
    elif optimizer_type.lower() == "adamax":
        return torch.optim.Adamax(model.parameters(),
                                  lr=learning_rate,
                                  weight_decay=kwargs.get("weight_decay", 0))
    elif optimizer_type.lower() == "rmsprop":
        return torch.optim.RMSprop(model.parameters(),
                                   lr=learning_rate,
                                   momentum=kwargs.get("momentum", 0.9))
    elif optimizer_type.lower() == "rprop":
        return torch.optim.Rprop(model.parameters(),
                                 lr=learning_rate)
    elif optimizer_type.lower() == "sgd":
        return torch.optim.SGD(model.parameters(),
                               lr=learning_rate,
                               momentum=kwargs.get("momentum", 0.9))



import numpy as np

from petrel.model import load_edet, load_optimizer, load_scheduler

from unittest import TestCase


class MyTestCase(TestCase):
    def test_scheduler(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("adam", model, learning_rate=1e-2)
        scheduler = load_scheduler("exponential", optimizer=optimizer)
        assert scheduler.gamma == 1.0
        assert scheduler.get_last_lr()[0] == 1e-2
        optimizer.step()
        scheduler.step()
        assert scheduler.get_last_lr()[0] == 1e-2

    def test_exponential(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("adam", model, learning_rate=1e-2)
        scheduler = load_scheduler("exponential", optimizer=optimizer, gamma=0.9)
        assert scheduler.gamma == 0.9
        assert scheduler.get_last_lr()[0] == 1e-2
        optimizer.step()
        scheduler.step()
        assert np.round(scheduler.get_last_lr()[0], 3) == 9e-3
        optimizer.step()
        scheduler.step()
        assert np.round(scheduler.get_last_lr()[0], 4) == 8.1e-3

    def test_mult(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("adam", model, learning_rate=1e-2)
        scheduler = load_scheduler("mult", optimizer=optimizer, lr_lambda=lambda epoch: 0.9)
        assert scheduler.get_last_lr()[0] == 1e-2
        optimizer.step()
        scheduler.step()
        assert np.round(scheduler.get_last_lr()[0], 3) == 9e-3
        optimizer.step()
        scheduler.step()
        assert np.round(scheduler.get_last_lr()[0], 4) == 8.1e-3

    def test_cosine(self):
        model = load_edet("tf_efficientdet_d0", image_size=512)
        optimizer = load_optimizer("adam", model, learning_rate=1e-2)
        scheduler = load_scheduler("cosine",
                                   optimizer=optimizer,
                                   T_max=100,
                                   eta_min=0)
        assert scheduler.get_last_lr()[0] == 1e-2
        optimizer.step()
        scheduler.step()
        assert scheduler.get_last_lr()[0] == 0.009997532801828659
        optimizer.step()
        scheduler.step()
        assert np.round(scheduler.get_last_lr()[0], 7) == 0.0099901